#!/bin/sh

#
# This runit service emulates the normal Docker ENTRYPOINT/CMD handling,
# insofar as possible.
#
# Docker's semantics for ENTRYPOINT/CMD vary depending on if those
# values are specified as arrays of string, or simple as single strings
# (which must be interpreted by /bin/sh -c).  We mark (inside
# /etc/emulab/docker) the type of entrypoint and cmd fields built into
# the image, so that the shell will know what to do.  Then at runtime,
# if the user changes entrypoint/cmd for any particular container, we
# mark the types of those fields, too.  We drop the image content into
# simple text files (entrypoint.image and entrypoint.image.type,
# cmd.image and cmd.image.type) as shell commands or arguments, and we
# drop the dynamic content into entrypoint.runtime{.type} and
# cmd.runtime{.type} as bind mounts.
#
# Given the presence (or absence!) of those files, this script
# implements the emulation, based upon the content in those files:
#
# if entrypoint.runtime.type == string:
#   Run exactly the command in entrypoint.runtime
# elif entrypoint.image.type == string:
#   Run exactly the command in entrypoint.image
# else:
#   cmd = ""
#   if entrypoint.runtime != "":
#     cmd = `cat entrypoint.runtime`
#   elif entrypoint.image != "":
#     cmd = `cat entrypoint.image`
#   if cmd.runtime.type == string:
#     strings = `cat cmd.runtime`
#     cmd = "$cmd /bin/sh -c $strings"
#   elif cmd.image.type == string:
#     strings = `cat cmd.image`
#     cmd = "$cmd /bin/sh -c $strings"
#   elif -n cmd.runtime:
#     strings = `cat cmd.runtime`
#     cmd = "$cmd $strings"
#   elif -n cmd.image:
#     strings = `cat cmd.image`
#     cmd = "$cmd $strings"
#
# If we still have nothing to run, we down the service and exit.
#
# Before executing "$cmd", we include the dockerenv.image file, then
# include the dockerenv.runtime file, if either exists.  Finally, we check
# to see if a USER was specified for the image; and if so, exec "$cmd"
# as that USER via chpst.
#
# When we run chpst, we also close stdin, and we redirect $cmd's outputs
# to /var/log/entrypoint.log.
#

PREFIX=/etc/emulab/docker
EPI=""
EPI_TYPE=""
if [ -e $PREFIX/entrypoint.image ]; then
    EPI=`cat $PREFIX/entrypoint.image`
fi
if [ -e $PREFIX/entrypoint.image.type ]; then
    EPI_TYPE=`cat $PREFIX/entrypoint.image.type`
fi
EPR=""
EPR_TYPE=""
if [ -e $PREFIX/entrypoint.runtime ]; then
    EPR=`cat $PREFIX/entrypoint.runtime`
fi
if [ -e $PREFIX/entrypoint.runtime.type ]; then
    EPR_TYPE=`cat $PREFIX/entrypoint.runtime.type`
fi
CMDI=""
CMDI_TYPE=""
if [ -e $PREFIX/cmd.image ]; then
    CMDI=`cat $PREFIX/cmd.image`
fi
if [ -e $PREFIX/cmd.image.type ]; then
    CMDI_TYPE=`cat $PREFIX/cmd.image.type`
fi
CMDR=""
CMDR_TYPE=""
if [ -e $PREFIX/cmd.runtime ]; then
    CMDR=`cat $PREFIX/cmd.runtime`
fi
if [ -e $PREFIX/cmd.runtime.type ]; then
    CMDR_TYPE=`cat $PREFIX/cmd.runtime.type`
fi

USER=""
if [ -e $PREFIX/user ]; then
    USER=`cat $PREFIX/user`
fi
WORKINGDIR=""
if [ -e $PREFIX/workingdir ]; then
    WORKINGDIR=`cat $PREFIX/workingdir`
fi
if [ -z "$WORKINGDIR" ]; then
    WORKINGDIR="/"
fi
ENVFILE_G=$PREFIX/dockerenv.generated
ENVFILE_I=$PREFIX/dockerenv.image
ENVFILE_R=$PREFIX/dockerenv.runtime

# Set up the cmd string.
cmd=""
EPT=""
if [ -n "$EPR_TYPE" -a "$EPR_TYPE" = "string" ]; then
    cmd="/bin/sh -c $EPR"
elif  [ -n "$EPI_TYPE" -a "$EPI_TYPE" = "string" ]; then
    cmd="/bin/sh -c $EPI"
else
    if [ -n "$EPR" ]; then
	cmd="$EPR"
	EPT=${EPR_TYPE}
    elif [ -n "$EPI" ]; then
	cmd="$EPI"
	EPT=${EPI_TYPE}
    fi
    if [ -n "${CMDR_TYPE}" -a "${CMDR_TYPE}" = "string" ]; then
	if [ "${CMDR_TYPE}" = "string" -a \( -z "$EPT" -o "$EPT" = "array" \) ]; then
	    cmd="$cmd /bin/sh -c"
	fi
	cmd="$cmd $CMDR"
    elif [ -n "${CMDI_TYPE}" -a "${CMDI_TYPE}" = "string" ]; then
	if [ "${CMDI_TYPE}" = "string" -a \( -z "$EPT" -o "$EPT" = "array" \) ]; then
	    cmd="$cmd /bin/sh -c"
	fi
	cmd="$cmd $CMDI"
    elif [ -n "$CMDR" ]; then
	cmd="$cmd $CMDR"
    elif [ -n "$CMDI" ]; then
	cmd="$cmd $CMDI"
    fi
fi

if [ -z "$cmd" ]; then
    # The finish script will down us, to emulate a oneshot service.
    exit 0
fi

chpst="/usr/bin/chpst -0"
if [ -n "$USER" ]; then
    chpst="$chpst -U $USER"
fi

if [ -e $ENVFILE_G ]; then
    . $ENVFILE_G
fi
if [ -e $ENVFILE_I ]; then
    . $ENVFILE_I
fi
if [ -e $ENVFILE_R ]; then
    . $ENVFILE_R
fi

exec 2>&1
mkdir /var
mkdir /var/log
exec > /var/log/entrypoint.log

date
env
echo Running $chpst $cmd

cd $WORKINGDIR
exec "$chpst $cmd" || exit 999
